# :coding: utf-8

import pytest
import types
import re

import six.moves

from wiz.utility import Requirement, Version
import wiz.graph
import wiz.package
import wiz.definition
import wiz.exception


@pytest.fixture()
def mocked_queue(mocker):
    """Return mocked Queue constructor."""
    instance = mocker.Mock()
    try:
        import queue
        mocker.patch("queue.Queue", return_value=instance)

    except ImportError:
        import Queue
        mocker.patch("Queue.Queue", return_value=instance)

    return instance


@pytest.fixture()
def mocked_graph(mocker):
    """Return mocked Graph."""
    graph = mocker.patch.object(wiz.graph, "Graph")
    graph.ROOT = "root"
    return graph


@pytest.fixture()
def mocked_resolver(mocker):
    """Return mocked Resolver."""
    resolver = mocker.patch.object(wiz.graph, "Resolver")
    return resolver


@pytest.fixture()
def mocked_graph_remove_node(mocker):
    """Return mocked Graph.remove_node method."""
    return mocker.patch.object(wiz.graph.Graph, "remove_node")


@pytest.fixture()
def mocked_updated_by_distance(mocker):
    """Return mocked updated_by_distance method."""
    return mocker.patch.object(wiz.graph, "updated_by_distance")


@pytest.fixture()
def mocked_package_extract(mocker):
    """Return mocked wiz.package.extract method."""
    return mocker.patch.object(wiz.package, "extract")


def test_resolver():
    """Create a resolver."""
    definition_mapping = {"defA": ["nodeA"], "defB": ["nodeB"]}
    resolver = wiz.graph.Resolver(definition_mapping)
    assert resolver.definition_mapping == definition_mapping
    assert id(resolver.definition_mapping) == id(definition_mapping)


@pytest.mark.parametrize("mapping, expected", [
    (
        {"root": []},
        {"root": {"distance": 0, "parent": "root"}}
    ),
    (
        {"root": ["A"], "A": []},
        {
            "root": {"distance": 0, "parent": "root"},
            "A": {"distance": 1, "parent": "root"}
        }
    ),
    (
        {"root": ["A", "B"], "A": [], "B": []},
        {
            "root": {"distance": 0, "parent": "root"},
            "A": {"distance": 1, "parent": "root"},
            "B": {"distance": 2, "parent": "root"}
        }
    ),
    (
        {"root": ["A", "B", "C"], "A": [], "B": [], "C": []},
        {
            "root": {"distance": 0, "parent": "root"},
            "A": {"distance": 1, "parent": "root"},
            "B": {"distance": 2, "parent": "root"},
            "C": {"distance": 3, "parent": "root"}
        }
    ),
    (
        {"root": ["A"], "A": ["B"], "B": []},
        {
            "root": {"distance": 0, "parent": "root"},
            "A": {"distance": 1, "parent": "root"},
            "B": {"distance": 2, "parent": "A"}
        }
    ),
    (
        {"root": ["A"], "A": ["B"], "B": ["C"], "C": []},
        {
            "root": {"distance": 0, "parent": "root"},
            "A": {"distance": 1, "parent": "root"},
            "B": {"distance": 2, "parent": "A"},
            "C": {"distance": 3, "parent": "B"}
        }
    ),
    (
        {"root": ["A", "B"], "A": ["B"], "B": ["A"]},
        {
            "root": {"distance": 0, "parent": "root"},
            "A": {"distance": 1, "parent": "root"},
            "B": {"distance": 2, "parent": "root"}
        }
    ),
    (
        {
            "root": ["A", "B"],
            "A": ["C", "D"],
            "B": ["D", "F", "G"],
            "C": [],
            "D": ["E"],
            "E": [],
            "F": [],
            "G": []
        },
        {
            "root": {"distance": 0, "parent": "root"},
            "A": {"distance": 1, "parent": "root"},
            "B": {"distance": 2, "parent": "root"},
            "C": {"distance": 2, "parent": "A"},
            "D": {"distance": 3, "parent": "A"},
            "E": {"distance": 4, "parent": "D"},
            "F": {"distance": 4, "parent": "B"},
            "G": {"distance": 5, "parent": "B"},
        }
    ),
    (
        {
            "root": ["A"],
            "A": ["C", "E", "F"],
            "B": ["F", "D"],
            "C": [],
            "D": [],
            "E": [],
            "F": []
        },
        {
            "root": {"distance": 0, "parent": "root"},
            "A": {"distance": 1, "parent": "root"},
            "B": {"distance": None, "parent": None},
            "C": {"distance": 2, "parent": "A"},
            "D": {"distance": None, "parent": None},
            "E": {"distance": 3, "parent": "A"},
            "F": {"distance": 4, "parent": "A"},
        }

    ),
    (
        {
            "root": ["A", "B", "F"],
            "A": ["C", "D"],
            "B": ["D", "F", "G"],
            "C": ["G"],
            "D": ["E"],
            "E": ["A"],
            "F": [],
            "G": ["B"],
        },
        {
            "root": {"distance": 0, "parent": "root"},
            "A": {"distance": 1, "parent": "root"},
            "B": {"distance": 2, "parent": "root"},
            "C": {"distance": 2, "parent": "A"},
            "D": {"distance": 3, "parent": "A"},
            "E": {"distance": 4, "parent": "D"},
            "F": {"distance": 3, "parent": "root"},
            "G": {"distance": 3, "parent": "C"},
        }
    )
], ids=[
    "no-node",
    "one-node",
    "two-nodes",
    "three-nodes",
    "two-levels",
    "three-levels",
    "with-cycle",
    "multi-levels",
    "multi-levels-with-unreachable-nodes",
    "complex-multi-levels"
])
def test_compute_distance_mapping(mocker, mocked_graph, mapping, expected):
    """Compute distance mapping from Graph."""
    nodes = [mocker.Mock(identifier=_id) for _id in mapping.keys()]
    mocked_graph.nodes.return_value = nodes
    mocked_graph.outcoming = lambda x: mapping[x]
    mocked_graph.link_weight = lambda x, y: mapping[y].index(x) + 1
    assert wiz.graph.compute_distance_mapping(mocked_graph) == expected


@pytest.mark.parametrize("variant_groups, expected", [
    (
        [["foo[V2]", "foo[V1]"]],
        [("foo[V1]",), ("foo[V2]",)]
    ),
    (
        [["foo[V3]", "foo[V2]", "foo[V1]"]],
        [
            ("foo[V2]", "foo[V1]"),
            ("foo[V3]", "foo[V1]"),
            ("foo[V3]", "foo[V2]")
        ]
    ),
    (
        [["foo[V4]", "foo[V3]", "foo[V2]", "foo[V1]"]],
        [
            ("foo[V3]", "foo[V2]", "foo[V1]"),
            ("foo[V4]", "foo[V2]", "foo[V1]"),
            ("foo[V4]", "foo[V3]", "foo[V1]"),
            ("foo[V4]", "foo[V3]", "foo[V2]")
        ]
    ),
    (
        [["foo[V3]==1", "foo[V2]==1", "foo[V1]==1", "foo[V1]==2"]],
        [
            ("foo[V2]==1", "foo[V1]==1", "foo[V1]==2"),
            ("foo[V3]==1", "foo[V1]==1", "foo[V1]==2"),
            ("foo[V3]==1", "foo[V2]==1")
        ]
    ),
    (
        [["foo[V3]", "foo[V2]", "foo[V1]"], ["bar[V1]", "bar[V2]"]],
        [
            ("foo[V2]", "foo[V1]", "bar[V2]"),
            ("foo[V2]", "foo[V1]", "bar[V1]"),
            ("foo[V3]", "foo[V1]", "bar[V2]"),
            ("foo[V3]", "foo[V1]", "bar[V1]"),
            ("foo[V3]", "foo[V2]", "bar[V2]"),
            ("foo[V3]", "foo[V2]", "bar[V1]")
        ]
    ),
    (
        [
            ["foo[V3]", "foo[V2]", "foo[V1]"],
            ["bar[V4]", "bar[V3]", "bar[V2]", "bar[V1]"],
            ["bim[V2]", "bim[V1]"]
        ],
        [
            ("foo[V2]", "foo[V1]", "bar[V3]", "bar[V2]", "bar[V1]", "bim[V1]"),
            ("foo[V2]", "foo[V1]", "bar[V3]", "bar[V2]", "bar[V1]", "bim[V2]"),
            ("foo[V2]", "foo[V1]", "bar[V4]", "bar[V2]", "bar[V1]", "bim[V1]"),
            ("foo[V2]", "foo[V1]", "bar[V4]", "bar[V2]", "bar[V1]", "bim[V2]"),
            ("foo[V2]", "foo[V1]", "bar[V4]", "bar[V3]", "bar[V1]", "bim[V1]"),
            ("foo[V2]", "foo[V1]", "bar[V4]", "bar[V3]", "bar[V1]", "bim[V2]"),
            ("foo[V2]", "foo[V1]", "bar[V4]", "bar[V3]", "bar[V2]", "bim[V1]"),
            ("foo[V2]", "foo[V1]", "bar[V4]", "bar[V3]", "bar[V2]", "bim[V2]"),
            ("foo[V3]", "foo[V1]", "bar[V3]", "bar[V2]", "bar[V1]", "bim[V1]"),
            ("foo[V3]", "foo[V1]", "bar[V3]", "bar[V2]", "bar[V1]", "bim[V2]"),
            ("foo[V3]", "foo[V1]", "bar[V4]", "bar[V2]", "bar[V1]", "bim[V1]"),
            ("foo[V3]", "foo[V1]", "bar[V4]", "bar[V2]", "bar[V1]", "bim[V2]"),
            ("foo[V3]", "foo[V1]", "bar[V4]", "bar[V3]", "bar[V1]", "bim[V1]"),
            ("foo[V3]", "foo[V1]", "bar[V4]", "bar[V3]", "bar[V1]", "bim[V2]"),
            ("foo[V3]", "foo[V1]", "bar[V4]", "bar[V3]", "bar[V2]", "bim[V1]"),
            ("foo[V3]", "foo[V1]", "bar[V4]", "bar[V3]", "bar[V2]", "bim[V2]"),
            ("foo[V3]", "foo[V2]", "bar[V3]", "bar[V2]", "bar[V1]", "bim[V1]"),
            ("foo[V3]", "foo[V2]", "bar[V3]", "bar[V2]", "bar[V1]", "bim[V2]"),
            ("foo[V3]", "foo[V2]", "bar[V4]", "bar[V2]", "bar[V1]", "bim[V1]"),
            ("foo[V3]", "foo[V2]", "bar[V4]", "bar[V2]", "bar[V1]", "bim[V2]"),
            ("foo[V3]", "foo[V2]", "bar[V4]", "bar[V3]", "bar[V1]", "bim[V1]"),
            ("foo[V3]", "foo[V2]", "bar[V4]", "bar[V3]", "bar[V1]", "bim[V2]"),
            ("foo[V3]", "foo[V2]", "bar[V4]", "bar[V3]", "bar[V2]", "bim[V1]"),
            ("foo[V3]", "foo[V2]", "bar[V4]", "bar[V3]", "bar[V2]", "bim[V2]")
        ]
    ),
    (
        [
            ["A[V2]==1", "A[V1]==1", "A[V1]==2"],
            ["B[V3]==1", "B[V2]==1", "B[V1]==1", "B[V2]==2"]
        ],
        [
            ("A[V1]==1", "A[V1]==2", "B[V2]==1", "B[V1]==1", "B[V2]==2"),
            ("A[V1]==1", "A[V1]==2", "B[V3]==1", "B[V1]==1"),
            ("A[V1]==1", "A[V1]==2", "B[V3]==1", "B[V2]==1", "B[V2]==2"),
            ("A[V2]==1", "B[V2]==1", "B[V1]==1", "B[V2]==2"),
            ("A[V2]==1", "B[V3]==1", "B[V1]==1"),
            ("A[V2]==1", "B[V3]==1", "B[V2]==1", "B[V2]==2")
        ]
    ),
], ids=[
    "one-group-of-2",
    "one-group-of-3",
    "one-group-of-4",
    "one-group-with-conflicts",
    "two-groups",
    "three-groups",
    "multiple-groups-with-conflicts",
])
def test_generate_variant_combinations(
    mocker, mocked_graph, variant_groups, expected
):
    """Return list of node trimming combinations from variants."""
    # Suppose that variants are always between square brackets in identifier.
    mocked_graph.node = lambda _id: mocker.Mock(
        identifier=_id,
        package=mocker.Mock(
            variant_identifier=re.search(r"(?<=\[).+(?=\])", _id).group(0)
        )
    )

    results = wiz.graph.generate_variant_combinations(
        mocked_graph, variant_groups
    )
    assert isinstance(results, types.GeneratorType) is True
    for combination, _expected in six.moves.zip_longest(results, expected):
        assert combination[0] == mocked_graph
        assert combination[1] == _expected


def test_trim_unreachable_from_graph(
    mocker, mocked_graph, mocked_graph_remove_node
):
    """Remove unreachable nodes from graph based on distance mapping."""
    mocked_graph.nodes.return_value = [
        mocker.Mock(identifier="A"),
        mocker.Mock(identifier="B"),
        mocker.Mock(identifier="C"),
        mocker.Mock(identifier="D"),
        mocker.Mock(identifier="E"),
        mocker.Mock(identifier="F"),
    ]

    distance_mapping = {
        "root": {"distance": 0, "parent": "root"},
        "A": {"distance": 1, "parent": "root"},
        "B": {"distance": None, "parent": None},
        "C": {"distance": 2, "parent": "A"},
        "D": {"distance": None, "parent": None},
        "E": {"distance": 3, "parent": "A"},
        "F": {"distance": 4, "parent": "A"},
    }

    wiz.graph.trim_unreachable_from_graph(mocked_graph, distance_mapping)

    assert mocked_graph_remove_node.call_count == 2
    mocked_graph_remove_node.assert_any_call("B")
    mocked_graph_remove_node.assert_any_call("D")


@pytest.mark.parametrize("distance_mapping, nodes_removed", [
    (
        {
            "root": {"distance": 0, "parent": "root"},
            "A": {"distance": 1, "parent": "root"},
            "B": {"distance": 2, "parent": "root"},
            "C": {"distance": 3, "parent": "root"},
            "D": {"distance": None, "parent": None},
        },
        ["B"]
    ),
    (
        {
            "root": {"distance": 0, "parent": "root"},
            "A": {"distance": 1, "parent": "root"},
            "B": {"distance": 2, "parent": "root"},
            "C": {"distance": 3, "parent": "root"},
            "D": {"distance": 4, "parent": "A"},
        },
        []
    ),
], ids=[
    "true",
    "false",
])
def test_trim_invalid_conditions_from_graph(
    mocker, mocked_graph, mocked_graph_remove_node, distance_mapping,
    nodes_removed
):
    """Remove invalid nodes from graph based on distance mapping."""
    mocked_graph.find.side_effect = [
        ["B"], ["C"], ["C"], ["D"],
    ]

    mocked_graph.nodes.return_value = [
        mocker.Mock(
            identifier="A",
            package=mocker.Mock(conditions=["B", "C"]),
            parent_identifiers=[]

        ),
        mocker.Mock(
            identifier="B",
            package=mocker.Mock(conditions=["C", "D"]),
            parent_identifiers=[]
        ),
        mocker.Mock(
            identifier="C",
            package=mocker.Mock(conditions=[]),
            parent_identifiers=[]
        ),
        mocker.Mock(
            identifier="D",
            package=mocker.Mock(conditions=[]),
            parent_identifiers=[]
        ),
    ]

    result = wiz.graph.trim_invalid_from_graph(mocked_graph, distance_mapping)
    assert result == (len(nodes_removed) > 0)

    assert mocked_graph_remove_node.call_count == len(nodes_removed)
    for node in nodes_removed:
        mocked_graph_remove_node.assert_any_call(node)


def test_updated_by_distance():
    """Update nodes based on distance mapping."""
    identifiers = ["F", "E", "D", "C", "B", "A"]

    distance_mapping = {
        "root": {"distance": 0, "parent": "root"},
        "A": {"distance": 1, "parent": "root"},
        "B": {"distance": None, "parent": None},
        "C": {"distance": 2, "parent": "A"},
        "D": {"distance": None, "parent": None},
        "E": {"distance": 3, "parent": "A"},
        "F": {"distance": 4, "parent": "A"},
    }

    result = wiz.graph.updated_by_distance(identifiers, distance_mapping)
    assert result == ["A", "C", "E", "F"]


def test_extract_conflicting_nodes(mocker, mocked_graph):
    """Extract conflicting nodes for a specific node."""
    node_mapping = {
        "A": mocker.Mock(
            identifier="A",
            definition=mocker.Mock(qualified_identifier="defB")
        ),
        "B": mocker.Mock(
            identifier="B",
            definition=mocker.Mock(qualified_identifier="defB")
        ),
        "C": mocker.Mock(
            identifier="C",
            definition=mocker.Mock(qualified_identifier="defB")
        ),
        "D": mocker.Mock(
            identifier="D",
            definition=mocker.Mock(qualified_identifier="defA")
        ),
        "E": mocker.Mock(
            identifier="E",
            definition=mocker.Mock(qualified_identifier="defA")
        ),
        "F": mocker.Mock(
            identifier="F",
            definition=mocker.Mock(qualified_identifier="defA")
        )
    }

    mocked_graph.node = lambda _id: node_mapping[_id]
    mocked_graph.conflicting_identifiers.return_value = sorted(
        node_mapping.keys()
    )

    assert wiz.graph.extract_conflicting_nodes(
        mocked_graph, node_mapping["F"]
    ) == [node_mapping["D"], node_mapping["E"]]

    assert wiz.graph.extract_conflicting_nodes(
        mocked_graph, node_mapping["E"]
    ) == [node_mapping["D"], node_mapping["F"]]

    assert wiz.graph.extract_conflicting_nodes(
        mocked_graph, node_mapping["C"]
    ) == [node_mapping["A"], node_mapping["B"]]


def test_combined_requirements(mocker, mocked_graph):
    """Combine nodes requirements."""
    requirements = [
        Requirement("A >= 1"),
        Requirement("A >= 1, < 2"),
        Requirement("A == 1.2.3")
    ]

    nodes = [
        mocker.Mock(
            identifier="A==3",
            parent_identifiers=["B"]
        ),
        mocker.Mock(
            identifier="A==1.9",
            parent_identifiers=["C"]
        ),
        mocker.Mock(
            identifier="A==1.2.3",
            parent_identifiers=["D"]
        )
    ]

    mocked_graph.link_requirement.side_effect = requirements

    requirement = wiz.graph.combined_requirements(mocked_graph, nodes)

    assert str(requirement) == "A >=1, ==1.2.3, <2"

    assert mocked_graph.link_requirement.call_count == 3
    mocked_graph.link_requirement.assert_any_call("A==3", "B")
    mocked_graph.link_requirement.assert_any_call("A==1.9", "C")
    mocked_graph.link_requirement.assert_any_call("A==1.2.3", "D")


def test_combined_requirements_error(mocker, mocked_graph):
    """Fail to combine nodes requirements from different definition name."""
    requirements = [
        Requirement("A >= 1"),
        Requirement("Z >= 1, < 2"),
        Requirement("A == 1.2.3")
    ]

    nodes = [
        mocker.Mock(
            identifier="A==3",
            parent_identifiers=["B"]
        ),
        mocker.Mock(
            identifier="Z==1.9",
            parent_identifiers=["C"]
        ),
        mocker.Mock(
            identifier="A==1.2.3",
            parent_identifiers=["D"]
        )
    ]

    mocked_graph.link_requirement.side_effect = requirements

    with pytest.raises(wiz.exception.GraphResolutionError):
        wiz.graph.combined_requirements(mocked_graph, nodes)

    assert mocked_graph.link_requirement.call_count == 2
    mocked_graph.link_requirement.assert_any_call("A==3", "B")
    mocked_graph.link_requirement.assert_any_call("Z==1.9", "C")


@pytest.mark.parametrize("requirement, namespace, expected", [
    (
        Requirement("foo"),
        "namespace",
        Requirement("namespace::foo")
    ),
    (
        Requirement("namespace::foo"),
        "namespace",
        Requirement("namespace::foo")
    ),
    (
        Requirement("bar"),
        None,
        Requirement("bar")
    ),
    (
        Requirement("::bar"),
        None,
        Requirement("bar")
    ),
], ids=[
    "implicit-namespace",
    "explicit-namespace",
    "implicit-no-namespace",
    "explicit-no-namespace",
])
def test_sanitize_requirement(requirement, namespace, expected):
    """Mutate the requirement according to package namespaces."""
    wiz.graph.sanitize_requirement(requirement, namespace)
    assert requirement == expected


def test_extract_conflicting_requirements(mocker, mocked_graph):
    """Extract conflicting requirements from nodes."""
    nodes = [
        mocker.Mock(
            definition=mocker.Mock(qualified_identifier="foo"),
            identifier="foo==3.2.1",
            parent_identifiers=["E", "F"]
        ),
        mocker.Mock(
            definition=mocker.Mock(qualified_identifier="foo"),
            identifier="foo==4.0.0",
            parent_identifiers=["G", "H", "I"]
        ),
        mocker.Mock(
            definition=mocker.Mock(qualified_identifier="foo"),
            identifier="foo==3.0.0",
            parent_identifiers=["root"]
        )
    ]

    # calls for: E, F, G, H, I
    mocked_graph.exists.side_effect = [False, True, True, True, True]

    # calls for: F, G, H, I, root
    mocked_graph.link_requirement.side_effect = [
        Requirement("foo ==3.*"),
        Requirement("foo >=4, <5"),
        Requirement("foo >=4, <5"),
        Requirement("foo"),
        Requirement("foo==3.0.0"),
    ]

    conflicts = wiz.graph.extract_conflicting_requirements(mocked_graph, nodes)

    assert mocked_graph.exists.call_count == 5
    mocked_graph.exists.assert_any_call("E")
    mocked_graph.exists.assert_any_call("F")
    mocked_graph.exists.assert_any_call("G")
    mocked_graph.exists.assert_any_call("H")
    mocked_graph.exists.assert_any_call("I")

    assert mocked_graph.link_requirement.call_count == 5
    mocked_graph.link_requirement.assert_any_call("foo==3.2.1", "F")
    mocked_graph.link_requirement.assert_any_call("foo==4.0.0", "G")
    mocked_graph.link_requirement.assert_any_call("foo==4.0.0", "H")
    mocked_graph.link_requirement.assert_any_call("foo==4.0.0", "I")
    mocked_graph.link_requirement.assert_any_call("foo==3.0.0", "root")

    assert conflicts == [
        {
            "graph": mocked_graph,
            "requirement": Requirement("foo >=4, <5"),
            "identifiers": {"G", "H"},
            "conflicts": {"F", "root"}
        },
        {
            "graph": mocked_graph,
            "requirement": Requirement("foo ==3.0.0"),
            "identifiers": {"root"},
            "conflicts": {"G", "H"}
        },
        {
            "graph": mocked_graph,
            "requirement": Requirement("foo ==3.*"),
            "identifiers": {"F"},
            "conflicts": {"G", "H"}
        }
    ]


def test_extract_conflicting_requirements_error(mocker, mocked_graph):
    """Fail to extract conflicting requirements from nodes."""
    nodes = [
        mocker.Mock(
            definition=mocker.Mock(qualified_identifier="bar"),
            identifier="bar==3.2.1",
            parent_identifiers=["E", "F"]
        ),
        mocker.Mock(
            definition=mocker.Mock(qualified_identifier="foo"),
            identifier="foo==4.0.0",
            parent_identifiers=["G", "H", "I"]
        ),
        mocker.Mock(
            definition=mocker.Mock(qualified_identifier="foo"),
            identifier="foo==3.0.0",
            parent_identifiers=["root"]
        )
    ]

    with pytest.raises(wiz.exception.GraphResolutionError) as error:
        wiz.graph.extract_conflicting_requirements(mocked_graph, nodes)

    assert (
        "All nodes should have the same definition identifier when "
        "attempting to extract conflicting requirements from parent "
        "nodes [bar, foo]"
    ) in str(error.value)

    mocked_graph.exists.assert_not_called()
    mocked_graph.link_requirement.assert_not_called()


def test_relink_parents(mocker, mocked_graph):
    """Relink node's parents."""
    node = mocker.Mock(
        identifier="foo",
        parent_identifiers=["parent1", "parent2", "parent3"]
    )

    mocked_graph.find.side_effect = [
        ["bar", "baz"], ["bim"]
    ]
    mocked_graph.exists.side_effect = [
        True, True, False,
        False,
        True, True
    ]
    mocked_graph.node.side_effect = [
        mocker.Mock(identifier="bar"),
        mocker.Mock(identifier="bim")
    ]
    mocked_graph.link_weight.side_effect = [1, 2]
    mocked_graph.link_requirement.side_effect = ["__REQ1__", "__REQ2__"]

    wiz.graph.relink_parents(mocked_graph, node)

    assert mocked_graph.find.call_count == 2
    mocked_graph.find.assert_any_call("__REQ1__")
    mocked_graph.find.assert_any_call("__REQ2__")

    assert mocked_graph.exists.call_count == 6
    mocked_graph.exists.assert_any_call("bar")
    mocked_graph.exists.assert_any_call("baz")
    mocked_graph.exists.assert_any_call("bim")
    mocked_graph.exists.assert_any_call("parent1")
    mocked_graph.exists.assert_any_call("parent2")
    mocked_graph.exists.assert_any_call("parent3")

    assert mocked_graph.link_weight.call_count == 2
    mocked_graph.link_weight.assert_any_call("foo", "parent1")
    mocked_graph.link_weight.assert_any_call("foo", "parent3")

    assert mocked_graph.link_requirement.call_count == 2
    mocked_graph.link_requirement.assert_any_call("foo", "parent1")
    mocked_graph.link_requirement.assert_any_call("foo", "parent3")

    assert mocked_graph.create_link.call_count == 2
    mocked_graph.create_link.assert_any_call(
        "bar", "parent1", "__REQ1__", weight=1
    )
    mocked_graph.create_link.assert_any_call(
        "bim", "parent3", "__REQ2__", weight=2
    )


def test_relink_parents_error(mocker, mocked_graph):
    """Fail to relink node's parents."""
    node = mocker.Mock(
        identifier="foo",
        parent_identifiers=["parent1", "parent2", "parent3"]
    )

    mocked_graph.find.side_effect = [
        ["bar", "baz"], ["bim"]
    ]
    mocked_graph.exists.side_effect = [
        True, True, False,
        False,
        True, False
    ]
    mocked_graph.node.side_effect = [
        mocker.Mock(identifier="bar")
    ]
    mocked_graph.link_weight.side_effect = [1, 2]
    mocked_graph.link_requirement.side_effect = ["__REQ1__", "__REQ2__"]

    with pytest.raises(wiz.exception.GraphResolutionError) as error:
        wiz.graph.relink_parents(mocked_graph, node)

    assert (
        "GraphResolutionError: 'parent3' can not be linked to any existing "
        "node in graph with requirement '__REQ2__'"
    ) in str(error)

    assert mocked_graph.find.call_count == 2
    mocked_graph.find.assert_any_call("__REQ1__")
    mocked_graph.find.assert_any_call("__REQ2__")

    assert mocked_graph.exists.call_count == 6
    mocked_graph.exists.assert_any_call("bar")
    mocked_graph.exists.assert_any_call("baz")
    mocked_graph.exists.assert_any_call("bim")
    mocked_graph.exists.assert_any_call("parent1")
    mocked_graph.exists.assert_any_call("parent2")
    mocked_graph.exists.assert_any_call("parent3")

    assert mocked_graph.link_weight.call_count == 2
    mocked_graph.link_weight.assert_any_call("foo", "parent1")
    mocked_graph.link_weight.assert_any_call("foo", "parent3")

    assert mocked_graph.link_requirement.call_count == 2
    mocked_graph.link_requirement.assert_any_call("foo", "parent1")
    mocked_graph.link_requirement.assert_any_call("foo", "parent3")

    assert mocked_graph.create_link.call_count == 1
    mocked_graph.create_link.assert_any_call(
        "bar", "parent1", "__REQ1__", weight=1
    )


def test_relink_parents_with_requirement(mocker, mocked_graph):
    """Relink node's parents with global requirement."""
    node = mocker.Mock(
        identifier="foo",
        parent_identifiers=["parent1", "parent2", "parent3"]
    )

    mocked_graph.find.return_value = ["bar", "baz", "bim"]
    mocked_graph.exists.side_effect = [
        True, False, True,
        True, False, True
    ]
    mocked_graph.node.side_effect = [
        mocker.Mock(identifier="bar"),
        mocker.Mock(identifier="bim")
    ]
    mocked_graph.link_weight.side_effect = [1, 2]

    wiz.graph.relink_parents(mocked_graph, node, "__REQUIREMENT__")

    assert mocked_graph.find.call_count == 1
    mocked_graph.find.assert_any_call("__REQUIREMENT__")

    assert mocked_graph.exists.call_count == 6
    mocked_graph.exists.assert_any_call("bar")
    mocked_graph.exists.assert_any_call("baz")
    mocked_graph.exists.assert_any_call("bim")
    mocked_graph.exists.assert_any_call("parent1")
    mocked_graph.exists.assert_any_call("parent2")
    mocked_graph.exists.assert_any_call("parent3")

    assert mocked_graph.link_weight.call_count == 2
    mocked_graph.link_weight.assert_any_call("foo", "parent1")
    mocked_graph.link_weight.assert_any_call("foo", "parent3")

    mocked_graph.link_requirement.assert_not_called()

    assert mocked_graph.create_link.call_count == 4
    mocked_graph.create_link.assert_any_call(
        "bar", "parent1", "__REQUIREMENT__", weight=1
    )
    mocked_graph.create_link.assert_any_call(
        "bim", "parent1", "__REQUIREMENT__", weight=1
    )
    mocked_graph.create_link.assert_any_call(
        "bar", "parent3", "__REQUIREMENT__", weight=2
    )
    mocked_graph.create_link.assert_any_call(
        "bim", "parent3", "__REQUIREMENT__", weight=2
    )


def test_validate_success(mocked_graph, mocked_updated_by_distance):
    """Validate graph with remaining errors."""
    mocked_graph.error_identifiers.return_value = []
    mocked_updated_by_distance.return_value = ["A", "B"]

    assert wiz.graph.validate(mocked_graph, "__DISTANCE_MAPPING__") is None

    mocked_updated_by_distance.assert_not_called()
    mocked_graph.errors.assert_not_called()


def test_validate_error(mocked_graph, mocked_updated_by_distance):
    """Validate graph with remaining errors."""
    error_mapping = {
        "A": [Exception("Error1"), Exception("Error2")],
        "B": [Exception("Error3")]
    }

    mocked_graph.error_identifiers.return_value = ["__ERRORS__"]
    mocked_graph.errors = lambda _id: error_mapping[_id]
    mocked_updated_by_distance.return_value = ["A", "B"]

    with pytest.raises(wiz.exception.GraphResolutionError) as error:
        wiz.graph.validate(mocked_graph, "__DISTANCE_MAPPING__")

    mocked_updated_by_distance.assert_called_once_with(
        ["__ERRORS__"], "__DISTANCE_MAPPING__"
    )

    assert (
       "The dependency graph could not be resolved due to the following "
       "error(s):\n"
       "  * A: Error1\n"
       "  * A: Error2\n"
    ) in str(error.value)


def test_validate_empty(mocked_graph, mocked_updated_by_distance):
    """Validate graph with remaining errors."""
    mocked_graph.error_identifiers.return_value = ["__ERRORS__"]
    mocked_updated_by_distance.return_value = []

    with pytest.raises(Exception) as error:
        wiz.graph.validate(mocked_graph, "__DISTANCE_MAPPING__")

    mocked_updated_by_distance.assert_called_once_with(
        ["__ERRORS__"], "__DISTANCE_MAPPING__"
    )
    mocked_graph.errors.assert_not_called()

    assert (
        "The dependency graph does not contain any valid packages."
        in str(error)
    )


@pytest.mark.parametrize("identifiers, distance_mapping, expected", [
    (
        [],
        {"root": {"distance": 0, "parent": "root"}},
        []
    ),
    (
        ["A"],
        {
            "root": {"distance": 0, "parent": "root"},
            "A": {"distance": 1, "parent": "root"}
        },
        ["A"]
    ),
    (
        ["A", "B"],
        {
            "root": {"distance": 0, "parent": "root"},
            "A": {"distance": 1, "parent": "root"},
            "B": {"distance": 2, "parent": "root"}
        },
        ["B", "A"],
    ),
    (
        ["A", "B", "C"],
        {
            "root": {"distance": 0, "parent": "root"},
            "A": {"distance": 1, "parent": "root"},
            "B": {"distance": 2, "parent": "root"},
            "C": {"distance": 3, "parent": "root"}
        },
        ["C", "B", "A"]
    ),
    (
        ["A", "B", "C", "D", "E", "F", "G"],
        {
            "root": {"distance": 0, "parent": "root"},
            "A": {"distance": 1, "parent": "root"},
            "B": {"distance": 2, "parent": "root"},
            "C": {"distance": 2, "parent": "A"},
            "D": {"distance": 3, "parent": "A"},
            "E": {"distance": 4, "parent": "D"},
            "F": {"distance": 4, "parent": "B"},
            "G": {"distance": 5, "parent": "B"},
        },
        ["G", "E", "F", "D", "B", "C", "A"]
    ),
    (
        ["A", "B", "C", "D", "E"],
        {
            "root": {"distance": 0, "parent": "root"},
            "A": {"distance": 1, "parent": "root"},
            "B": {"distance": 2, "parent": "root"},
            "C": {"distance": 2, "parent": "A"},
            "D": {"distance": None, "parent": "F"},
            "E": {"distance": None, "parent": "G"},
        },
        ["B", "C", "A"]
    )
], ids=[
    "no-node",
    "one-node",
    "two-nodes",
    "three-nodes",
    "multi-levels",
    "unreachable-nodes",
])
def test_extract_ordered_packages(
    mocker, mocked_graph, identifiers, distance_mapping, expected
):
    """Extract ordered packages from graph."""
    package_mapping = {
        _id: wiz.package.Package(
            wiz.definition.Definition({"identifier": _id})
        ) for _id in identifiers
    }

    nodes = [
        mocker.Mock(identifier=_id, package=package_mapping[_id])
        for _id in identifiers
    ]
    mocked_graph.nodes.return_value = nodes

    result = wiz.graph.extract_ordered_packages(mocked_graph, distance_mapping)
    assert result == [package_mapping[_id] for _id in expected]


def test_graph_node():
    """Return node from graph via identifier."""
    graph = wiz.graph.Graph(None)
    graph._node_mapping = {"A": "_nodeA"}

    assert graph.node("A") == "_nodeA"
    assert graph.node("B") is None


def test_graph_nodes():
    """Return nodes from graph."""
    graph = wiz.graph.Graph(None)
    graph._node_mapping = {"A": "_nodeA", "B": "_nodeB", "C": "_nodeC"}

    assert sorted(graph.nodes()) == ["_nodeA", "_nodeB", "_nodeC"]


def test_graph_exists():
    """Indicate whether node exists in graph."""
    graph = wiz.graph.Graph(None)
    graph._node_mapping = {"A": "_nodeA"}

    assert graph.exists("A") is True
    assert graph.exists("B") is False


@pytest.mark.parametrize("requirement, expected", [
    (
        Requirement("Z"),
        []
    ),
    (
        Requirement("A"),
        ["A==0.1.0", "A==2.4.5"]
    ),
    (
        Requirement("Name1::B"),
        ["Name1::B"]
    ),
    (
        Requirement("A > 1"),
        ["A==2.4.5"]
    )
], ids=[
    "un-found",
    "match-definition-name",
    "match-namespace",
    "match-versions"
])
def test_graph_find_matching_identifiers(
    requirement, expected, mocker
):
    """Find matching identifiers from requirement."""
    graph = wiz.graph.Graph(None)
    graph._node_mapping = {
        "A==0.1.0": mocker.Mock(
            definition=mocker.Mock(qualified_identifier="A"),
            package=mocker.Mock(
                version=Version("0.1.0"),
                identifier="A==0.1.0"
            )
        ),
        "A==2.4.5": mocker.Mock(
            definition=mocker.Mock(qualified_identifier="A"),
            package=mocker.Mock(
                version=Version("2.4.5"),
                identifier="A==2.4.5"
            )
        ),
        "Name1::B": mocker.Mock(
            definition=mocker.Mock(qualified_identifier="Name1::B"),
            package=mocker.Mock(
                version=None,
                identifier="Name1::B"
            )
        ),
        "B==1": mocker.Mock(
            definition=mocker.Mock(qualified_identifier="B"),
            package=mocker.Mock(
                version=Version("1"),
                identifier="B==1"
            )
        ),
    }

    result = graph.find(requirement)
    assert result == expected


def test_graph_variant_groups(mocker):
    """Extract variants from graph."""
    graph = wiz.graph.Graph(None)
    assert graph.variant_groups() == []

    graph = wiz.graph.Graph(None)
    graph._node_mapping = {
        "A[V1]==0.1.0": mocker.Mock(
            identifier="A[V1]==0.1.0",
            package=mocker.Mock(
                variant_identifier="V1",
                version=Version("0.1.0")
            )
        ),
        "A[V2]==0.1.0": mocker.Mock(
            identifier="A[V2]==0.1.0",
            package=mocker.Mock(
                variant_identifier="V2",
                version=Version("0.1.0")
            )
        ),
        "B[V1]==0.1.0": mocker.Mock(
            identifier="B[V1]==0.1.0",
            package=mocker.Mock(
                variant_identifier="V1",
                version=Version("0.1.0")
            )
        ),
        "B[V1]==0.2.0": mocker.Mock(
            identifier="B[V1]==0.2.0",
            package=mocker.Mock(
                variant_identifier="V1",
                version=Version("0.2.0")
            )
        ),
        "B[V2]==0.1.0": mocker.Mock(
            identifier="B[V2]==0.1.0",
            package=mocker.Mock(
                variant_identifier="V2",
                version=Version("0.1.0")
            )
        ),
        "C[V1]==0.1.0": mocker.Mock(
            identifier="C[V1]==0.1.0",
            package=mocker.Mock(
                variant_identifier="V1",
                version=Version("0.1.0")
            )
        ),
        "D[V1]==0.1.0": mocker.Mock(
            identifier="D[V1]==0.1.0",
            package=mocker.Mock(
                variant_identifier="V1",
                version=Version("0.1.0")
            )
        ),
        "D[V1]==0.2.0": mocker.Mock(
            identifier="D[V1]==0.2.0",
            package=mocker.Mock(
                variant_identifier="V1",
                version=Version("0.2.0")
            )
        )
    }
    graph._variants_per_definition = {
        "A": ["A[V1]==0.1.0", "A[V2]==0.1.0", "A[V2]==0.1.0"],
        "B": ["B[V1]==0.1.0", "B[V2]==0.1.0", "B[V1]==0.2.0"],
        "C": ["C[V1]==0.1.0"],
        "D": ["D[V1]==0.1.0", "D[V1]==0.2.0"]
    }
    assert graph.variant_groups() == [
        ["A[V2]==0.1.0", "A[V1]==0.1.0"],
        ["B[V1]==0.2.0", "B[V1]==0.1.0", "B[V2]==0.1.0"]
    ]


def test_graph_outcoming():
    """Fetch outcoming nodes from node identifier."""
    graph = wiz.graph.Graph(None)
    assert graph.outcoming("A") == []

    graph = wiz.graph.Graph(None)
    graph._link_mapping = {"A": {"B": "LINK"}}
    assert graph.outcoming("A") == []

    graph = wiz.graph.Graph(None)
    graph._node_mapping = {"A": "_A"}
    graph._link_mapping = {"A": {"B": "LINK"}}
    assert graph.outcoming("A") == []

    graph = wiz.graph.Graph(None)
    graph._node_mapping = {"A": "_A", "B": "_B"}
    graph._link_mapping = {"A": {"B": "LINK"}}
    assert graph.outcoming("A") == ["B"]


def test_graph_link_weight():
    """Fetch weight from nodes link."""
    graph = wiz.graph.Graph(None)
    graph._link_mapping = {
        "A": {
            "B": {"requirement": Requirement("A"), "weight": 3},
            "C": {"requirement": Requirement("A>2"), "weight": 2}
        }
    }
    assert graph.link_weight("B", "A") == 3
    assert graph.link_weight("C", "A") == 2


def test_graph_requirement_weight():
    """Fetch requirement from nodes link."""
    requirements = [
        Requirement("A"), Requirement("A>2")
    ]

    graph = wiz.graph.Graph(None)
    graph._link_mapping = {
        "A": {
            "B": {"requirement": requirements[0], "weight": 3},
            "C": {"requirement": requirements[1], "weight": 2}
        }
    }

    assert graph.link_requirement("B", "A") == requirements[0]
    assert graph.link_requirement("C", "A") == requirements[1]


@pytest.mark.parametrize("definition_mapping, node_mapping, expected", [
    (
        {"defA": ["nodeA"], "defB": ["nodeB"]},
        {"nodeA": "_nodeA", "nodeB": "_nodeB"},
        []
    ),
    (
        {"defA": ["nodeA1", "nodeA2"], "defB": ["nodeB"]},
        {"nodeA1": "_nodeA1", "nodeA2": "_nodeA2", "nodeB": "_nodeB"},
        ["nodeA1", "nodeA2"]
    ),
    (
        {"defA": ["nodeA1", "nodeA2"], "defB": ["nodeB"]},
        {"nodeA1": "_nodeA1", "nodeB": "_nodeB"},
        []
    )
], ids=[
    "without-conflicts",
    "with-two-conflicts",
    "with-conflicting-node-removed"
])
def test_graph_conflicts(definition_mapping, node_mapping, expected):
    """Extract conflicting nodes from graph."""
    graph = wiz.graph.Graph(None)
    graph._node_mapping = node_mapping
    graph._identifiers_per_definition = definition_mapping

    assert graph.conflicting_identifiers() == expected


def test_graph_update_from_requirements(
    mocker, mocked_resolver, mocked_package_extract
):
    """Update graph from requirements."""
    graph = wiz.graph.Graph(mocked_resolver)

    _mapping = {
        "A==0.2.0": wiz.package.Package(
            wiz.definition.Definition({
                "identifier": "A",
                "version": "0.2.0"
            })
        ),
        "B==2.1.1": wiz.package.Package(
            wiz.definition.Definition({
                "identifier": "B",
                "version": "2.1.1"
            })
        ),
    }

    mocked_package_extract.side_effect = [
        [_mapping["A==0.2.0"]],
        [_mapping["B==2.1.1"]]
    ]

    graph.update_from_requirements(
        [Requirement("A"), Requirement("B>=2")], graph.ROOT
    )

    assert graph.data() == {
        "identifier": mocker.ANY,
        "node_mapping": {
            "A==0.2.0": {
                "package": {
                    "identifier": "A==0.2.0",
                    "version": "0.2.0"
                },
                "parents": ["root"]
            },
            "B==2.1.1": {
                "package": {
                    "identifier": "B==2.1.1",
                    "version": "2.1.1"
                },
                "parents": ["root"]
            }
        },
        "link_mapping": {
            "root": {
                "A==0.2.0": {"requirement": Requirement("A"), "weight": 1},
                "B==2.1.1": {"requirement": Requirement("B >=2"), "weight": 2}
            }
        },
        "identifiers_per_definition": {
            "A": ["A==0.2.0"],
            "B": ["B==2.1.1"]
        },
        "variants_per_definition": {},
        "conditioned_nodes": [],
        "namespace_count": {},
        "error_mapping": {}
    }


def test_graph_update_from_requirements_with_dependencies(
    mocker, mocked_resolver, mocked_package_extract
):
    """Update graph from requirements with dependency requirements."""
    graph = wiz.graph.Graph(mocked_resolver)

    _mapping = {
        "A==0.1.0": wiz.package.Package(
            wiz.definition.Definition({
                "identifier": "A",
                "version": "0.1.0",
                "requirements": ["B>=2", "C"],
            })
        ),
        "B==3.0.0": wiz.package.Package(
            wiz.definition.Definition({
                "identifier": "B",
                "version": "3.0.0",
            })
        ),
        "C==1.2.3": wiz.package.Package(
            wiz.definition.Definition({
                "identifier": "C",
                "version": "1.2.3",
                "requirements": ["D"],
            })
        ),
        "D==0.1.0": wiz.package.Package(
            wiz.definition.Definition({
                "identifier": "D",
                "version": "0.1.0",
                "requirements": ["E"],
            })
        ),
        "E==0.2.0": wiz.package.Package(
            wiz.definition.Definition({
                "identifier": "E",
                "version": "0.2.0",
            })
        ),
    }

    mocked_package_extract.side_effect = [
        [_mapping["A==0.1.0"]],
        [_mapping["B==3.0.0"]],
        [_mapping["C==1.2.3"]],
        [_mapping["D==0.1.0"]],
        [_mapping["E==0.2.0"]],
    ]

    graph.update_from_requirements([Requirement("A")], graph.ROOT)

    assert graph.data() == {
        "identifier": mocker.ANY,
        "node_mapping": {
            "A==0.1.0": {
                "package": {
                    "identifier": "A==0.1.0",
                    "version": "0.1.0",
                    "requirements": ["B>=2", "C"]
                },
                "parents": ["root"]
            },
            "B==3.0.0": {
                "package": {
                    "identifier": "B==3.0.0",
                    "version": "3.0.0",
                },
                "parents": ["A==0.1.0"]
            },
            "C==1.2.3": {
                "package": {
                    "identifier": "C==1.2.3",
                    "version": "1.2.3",
                    "requirements": ["D"]
                },
                "parents": ["A==0.1.0"]
            },
            "D==0.1.0": {
                "package": {
                    "identifier": "D==0.1.0",
                    "version": "0.1.0",
                    "requirements": ["E"]
                },
                "parents": ["C==1.2.3"]
            },
            "E==0.2.0": {
                "package": {
                    "identifier": "E==0.2.0",
                    "version": "0.2.0",
                },
                "parents": ["D==0.1.0"]
            }
        },
        "link_mapping": {
            "root": {
                "A==0.1.0": {"requirement": Requirement("A"), "weight": 1}
            },
            "A==0.1.0": {
                "B==3.0.0": {"requirement": Requirement("B>=2"), "weight": 1},
                "C==1.2.3": {"requirement": Requirement("C"), "weight": 2}
            },
            "C==1.2.3": {
                "D==0.1.0": {"requirement": Requirement("D"), "weight": 1},
            },
            "D==0.1.0": {
                "E==0.2.0": {"requirement": Requirement("E"), "weight": 1},
            }
        },
        "identifiers_per_definition": {
            "A": ["A==0.1.0"],
            "B": ["B==3.0.0"],
            "C": ["C==1.2.3"],
            "D": ["D==0.1.0"],
            "E": ["E==0.2.0"]
        },
        "variants_per_definition": {},
        "conditioned_nodes": [],
        "namespace_count": {},
        "error_mapping": {}
    }


def test_graph_update_from_requirements_with_variants(
    mocker, mocked_resolver, mocked_package_extract
):
    """Update graph from requirements with variants of definition."""
    graph = wiz.graph.Graph(mocked_resolver)

    definition = wiz.definition.Definition({
        "identifier": "A",
        "version": "0.2.0",
        "variants": [
            {"identifier": "V1"},
            {"identifier": "V2"},
            {"identifier": "V3"},
        ]
    })

    _mapping = {
        "A[V1]==0.2.0": wiz.package.Package(definition, variant_index=0),
        "A[V2]==0.2.0": wiz.package.Package(definition, variant_index=1),
        "A[V3]==0.2.0": wiz.package.Package(definition, variant_index=2),
    }

    mocked_package_extract.side_effect = [
        [
            _mapping["A[V1]==0.2.0"],
            _mapping["A[V2]==0.2.0"],
            _mapping["A[V3]==0.2.0"]
        ],
    ]

    graph.update_from_requirements([Requirement("A")], graph.ROOT)

    assert graph.data() == {
        "identifier": mocker.ANY,
        "node_mapping": {
            "A[V1]==0.2.0": {
                "package": {
                    "identifier": "A[V1]==0.2.0",
                    "version": "0.2.0",
                    "variant-identifier": "V1",
                },
                "parents": ["root"]
            },
            "A[V2]==0.2.0": {
                "package": {
                    "identifier": "A[V2]==0.2.0",
                    "version": "0.2.0",
                    "variant-identifier": "V2",
                },
                "parents": ["root"]
            },
            "A[V3]==0.2.0": {
                "package": {
                    "identifier": "A[V3]==0.2.0",
                    "version": "0.2.0",
                    "variant-identifier": "V3",
                },
                "parents": ["root"]
            },
        },
        "link_mapping": {
            "root": {
                "A[V1]==0.2.0": {"requirement": Requirement("A"), "weight": 1},
                "A[V2]==0.2.0": {"requirement": Requirement("A"), "weight": 1},
                "A[V3]==0.2.0": {"requirement": Requirement("A"), "weight": 1}
            },
        },
        "identifiers_per_definition": {
            "A": ["A[V1]==0.2.0", "A[V2]==0.2.0", "A[V3]==0.2.0"],
        },
        "variants_per_definition": {
            "A": ["A[V1]==0.2.0", "A[V2]==0.2.0", "A[V3]==0.2.0"],
        },
        "conditioned_nodes": [],
        "namespace_count": {},
        "error_mapping": {}
    }


def test_graph_update_from_requirements_with_namespaces(
    mocker, mocked_resolver, mocked_package_extract
):
    """Update graph from requirements with namespaces."""
    mocked_resolver.definition_mapping = {
        "__namespace__": {
            "A": ["Foo"],
            "B": ["Foo", "Bar"],
        }
    }

    graph = wiz.graph.Graph(mocked_resolver)

    _mapping = {
        "Foo::A==0.2.0": wiz.package.Package(
            wiz.definition.Definition({
                "identifier": "A",
                "version": "0.2.0",
                "namespace": "Foo",
            })
        ),
        "Foo::B==2.1.1": wiz.package.Package(
            wiz.definition.Definition({
                "identifier": "B",
                "version": "2.1.1",
                "namespace": "Foo",
            })
        )
    }

    mocked_package_extract.side_effect = [
        [_mapping["Foo::A==0.2.0"]],
        [_mapping["Foo::B==2.1.1"]]
    ]

    graph.update_from_requirements(
        [Requirement("A"), Requirement("B>=2")], graph.ROOT
    )

    assert graph.data() == {
        "identifier": mocker.ANY,
        "node_mapping": {
            "Foo::A==0.2.0": {
                "package": {
                    "identifier": "Foo::A==0.2.0",
                    "namespace": "Foo",
                    "version": "0.2.0"
                },
                "parents": ["root"]
            },
            "Foo::B==2.1.1": {
                "package": {
                    "identifier": "Foo::B==2.1.1",
                    "namespace": "Foo",
                    "version": "2.1.1"
                },
                "parents": ["root"]
            }
        },
        "link_mapping": {
            "root": {
                "Foo::A==0.2.0": {
                    "requirement": Requirement("Foo::A"), "weight": 1
                },
                "Foo::B==2.1.1": {
                    "requirement": Requirement("Foo::B >=2"), "weight": 2
                }
            }
        },
        "identifiers_per_definition": {
            "Foo::A": ["Foo::A==0.2.0"],
            "Foo::B": ["Foo::B==2.1.1"]
        },
        "variants_per_definition": {},
        "conditioned_nodes": [],
        "namespace_count": {"Bar": 1, "Foo": 2},
        "error_mapping": {}
    }


def test_graph_update_from_requirements_with_skipped_conditional_packages(
    mocker, mocked_resolver, mocked_package_extract
):
    """Update graph from requirements with skipped conditional packages."""
    graph = wiz.graph.Graph(mocked_resolver)

    _mapping = {
        "A==0.2.0": wiz.package.Package(
            wiz.definition.Definition({
                "identifier": "A",
                "version": "0.2.0",
            })
        ),
        "B==2.1.1": wiz.package.Package(
            wiz.definition.Definition({
                "identifier": "B",
                "version": "2.1.1",
                "conditions": ["C > 2"]
            })
        ),
        "C==2.0.4": wiz.package.Package(
            wiz.definition.Definition({
                "identifier": "C",
                "version": "2.0.4",
            })
        ),
        "D==0.1.0": wiz.package.Package(
            wiz.definition.Definition({
                "identifier": "D",
                "version": "0.1.0",
                "conditions": ["W"]
            })
        ),
    }

    mocked_package_extract.side_effect = [
        [_mapping["A==0.2.0"]],
        [_mapping["B==2.1.1"]],
        [_mapping["D==0.1.0"]],
        # Extract conditional package
        [_mapping["C==2.0.4"]],
        wiz.exception.WizError("Oh Shit!")
    ]

    graph.update_from_requirements(
        [Requirement("A"), Requirement("B>=2"), Requirement("D")], graph.ROOT
    )

    assert graph.data() == {
        "identifier": mocker.ANY,
        "node_mapping": {
            "A==0.2.0": {
                "package": {
                    "identifier": "A==0.2.0",
                    "version": "0.2.0",
                },
                "parents": ["root"],
            }
        },
        "link_mapping": {
            "root": {
                "A==0.2.0": {"requirement": Requirement("A"), "weight": 1}
            }
        },
        "identifiers_per_definition": {
            "A": ["A==0.2.0"]
        },
        "conditioned_nodes": [
            {
                "requirement": Requirement("B >=2"),
                "package": {
                    "identifier": "B==2.1.1",
                    "version": "2.1.1",
                    "conditions": ["C > 2"]
                },
                "parent_identifier": "root",
                "weight": 2
            },
            {
                "requirement": Requirement("D"),
                "package": {
                    "identifier": "D==0.1.0",
                    "version": "0.1.0",
                    "conditions": ["W"]
                },
                "parent_identifier": "root",
                "weight": 3
            }
        ],
        "variants_per_definition": {},
        "namespace_count": {},
        "error_mapping": {}
    }


def test_graph_update_from_requirements_with_used_conditional_packages(
    mocker, mocked_resolver, mocked_package_extract
):
    """Update graph from requirements with used conditional packages."""
    graph = wiz.graph.Graph(mocked_resolver)

    _mapping = {
        "A==0.2.0": wiz.package.Package(
            wiz.definition.Definition({
                "identifier": "A",
                "version": "0.2.0"
            })
        ),
        "B==2.1.1": wiz.package.Package(
            wiz.definition.Definition({
                "identifier": "B",
                "version": "2.1.1",
                "conditions": ["A"],
            })
        )
    }

    mocked_package_extract.side_effect = [
        [_mapping["A==0.2.0"]],
        [_mapping["B==2.1.1"]],
        # Extract package from condition
        [_mapping["A==0.2.0"]]
    ]

    graph.update_from_requirements(
        [Requirement("A"), Requirement("B>=2")], graph.ROOT
    )

    assert graph.data() == {
        "identifier": mocker.ANY,
        "node_mapping": {
            "A==0.2.0": {
                "package": {
                    "identifier": "A==0.2.0",
                    "version": "0.2.0"
                },
                "parents": ["root"]
            },
            "B==2.1.1": {
                "package": {
                    "identifier": "B==2.1.1",
                    "version": "2.1.1",
                    "conditions": ["A"],
                },
                "parents": ["root"]
            }
        },
        "link_mapping": {
            "root": {
                "A==0.2.0": {"requirement": Requirement("A"), "weight": 1},
                "B==2.1.1": {"requirement": Requirement("B >=2"), "weight": 2}
            }
        },
        "identifiers_per_definition": {
            "A": ["A==0.2.0"],
            "B": ["B==2.1.1"]
        },
        "conditioned_nodes": [
            {
                "requirement": Requirement("B >=2"),
                "package": {
                    "identifier": "B==2.1.1",
                    "version": "2.1.1",
                    "conditions": ["A"],
                },
                "parent_identifier": "root",
                "weight": 2
            },
        ],
        "variants_per_definition": {},
        "namespace_count": {},
        "error_mapping": {}
    }


def test_graph_update_from_requirements_with_errors(
    mocker, mocked_resolver, mocked_package_extract
):
    """Update graph from requirements with errors."""
    graph = wiz.graph.Graph(mocked_resolver)

    _mapping = {
        "A==0.2.0": wiz.package.Package(
            wiz.definition.Definition({
                "identifier": "A",
                "version": "0.2.0"
            })
        ),
        "B==2.1.1": wiz.package.Package(
            wiz.definition.Definition({
                "identifier": "B",
                "version": "2.1.1",
                "requirements": ["incorrect1", "incorrect2"]
            })
        )
    }

    mocked_package_extract.side_effect = [
        [_mapping["A==0.2.0"]],
        [_mapping["B==2.1.1"]],
        wiz.exception.RequestNotFound("incorrect1"),
        wiz.exception.RequestNotFound("incorrect2"),
    ]

    graph.update_from_requirements(
        [Requirement("A"), Requirement("B>=2")], graph.ROOT
    )

    assert graph.data() == {
        "identifier": mocker.ANY,
        "node_mapping": {
            "A==0.2.0": {
                "package": {
                    "identifier": "A==0.2.0",
                    "version": "0.2.0"
                },
                "parents": ["root"]
            },
            "B==2.1.1": {
                "package": {
                    "identifier": "B==2.1.1",
                    "version": "2.1.1",
                    "requirements": ["incorrect1", "incorrect2"]
                },
                "parents": ["root"]
            }
        },
        "link_mapping": {
            "root": {
                "A==0.2.0": {"requirement": Requirement("A"), "weight": 1},
                "B==2.1.1": {"requirement": Requirement("B >=2"), "weight": 2}
            }
        },
        "identifiers_per_definition": {
            "A": ["A==0.2.0"],
            "B": ["B==2.1.1"]
        },
        "conditioned_nodes": [],
        "variants_per_definition": {},
        "namespace_count": {},
        "error_mapping": {
            "B==2.1.1": ["incorrect1", "incorrect2"],
        }
    }


@pytest.mark.parametrize("options", [
    {},
    {"weight": 5},
], ids=[
    "simple",
    "with-weight",
])
def test_graph_update_from_requirement_existing(
    mocker, mocked_resolver, mocked_package_extract, mocked_queue, options
):
    """Update graph from requirement."""
    package = wiz.package.Package(
        wiz.definition.Definition({
            "identifier": "A",
            "version": "0.1.0"
        })
    )
    node = mocker.Mock(identifier="_A==0.1.0")
    requirement = Requirement("A")

    mocked_package_extract.return_value = [package]

    graph = wiz.graph.Graph(mocked_resolver)
    graph.create_link = mocker.Mock()
    graph._create_node_from_package = mocker.Mock()
    graph._node_mapping = {"A==0.1.0": node}
    graph.exists = mocker.Mock(return_value=True)

    graph._update_from_requirement(
        requirement, graph.ROOT, mocked_queue, **options
    )

    graph._create_node_from_package.assert_not_called()
    graph.create_link.assert_called_once_with(
        "_A==0.1.0", graph.ROOT, requirement,
        weight=options.get("weight", 1)
    )

    node.add_parent.assert_called_once_with(graph.ROOT)

    mocked_queue.put.assert_not_called()


@pytest.mark.parametrize("options", [
    {},
    {"weight": 5},
], ids=[
    "simple",
    "with-weight",
])
def test_graph_update_from_requirement_non_existing(
    mocker, mocked_resolver, mocked_package_extract, mocked_queue, options
):
    """Update graph from requirement."""
    package = wiz.package.Package(
        wiz.definition.Definition({
            "identifier": "A",
            "version": "0.1.0"
        })
    )
    node = mocker.Mock(identifier="_A==0.1.0")
    requirement = Requirement("A")

    mocked_package_extract.return_value = [package]

    graph = wiz.graph.Graph(mocked_resolver)
    graph.create_link = mocker.Mock()
    graph._create_node_from_package = mocker.Mock()
    graph._node_mapping = {"A==0.1.0": node}
    graph.exists = mocker.Mock(return_value=False)

    graph._update_from_requirement(
        requirement, graph.ROOT, mocked_queue, **options
    )

    graph._create_node_from_package.assert_called_once_with(package)
    graph.create_link.assert_called_once_with(
        "_A==0.1.0",
        graph.ROOT,
        requirement,
        weight=options.get("weight", 1)
    )

    node.add_parent.assert_called_once_with(graph.ROOT)

    mocked_queue.put.assert_not_called()


@pytest.mark.parametrize("options", [
    {},
    {"weight": 5},
], ids=[
    "simple",
    "with-weight",
])
def test_graph_update_from_requirement_non_existing_with_requirements(
    mocker, mocked_resolver, mocked_package_extract, mocked_queue, options
):
    """Update graph from requirement."""
    package = wiz.package.Package(
        wiz.definition.Definition({
            "identifier": "A",
            "version": "0.1.0",
            "requirements": ["B", "C", "D"]
        })
    )
    node = mocker.Mock(identifier="_A==0.1.0")
    requirement = Requirement("A")

    mocked_package_extract.return_value = [package]

    graph = wiz.graph.Graph(mocked_resolver)
    graph.create_link = mocker.Mock()
    graph._create_node_from_package = mocker.Mock()
    graph._node_mapping = {"A==0.1.0": node}
    graph.exists = mocker.Mock(return_value=False)

    graph._update_from_requirement(
        requirement, graph.ROOT, mocked_queue, **options
    )

    graph._create_node_from_package.assert_called_once_with(package)
    graph.create_link.assert_called_once_with(
        "_A==0.1.0",
        graph.ROOT,
        requirement,
        weight=options.get("weight", 1)
    )

    node.add_parent.assert_called_once_with(graph.ROOT)

    assert mocked_queue.put.call_count == 3
    mocked_queue.put.assert_any_call({
        "requirement": Requirement("B"),
        "parent_identifier": "A==0.1.0",
        "weight": 1
    })
    mocked_queue.put.assert_any_call({
        "requirement": Requirement("C"),
        "parent_identifier": "A==0.1.0",
        "weight": 2
    })
    mocked_queue.put.assert_any_call({
        "requirement": Requirement("D"),
        "parent_identifier": "A==0.1.0",
        "weight": 3
    })


@pytest.mark.parametrize("options", [
    {},
    {"weight": 5},
], ids=[
    "simple",
    "with-weight",
])
def test_graph_update_from_requirement_multi_packages(
    mocker, mocked_resolver, mocked_package_extract, mocked_queue, options
):
    """Update graph from requirement."""
    definition = wiz.definition.Definition({
        "identifier": "A",
        "version": "0.1.0",
        "variants": [
            {"identifier": "variant1"},
            {"identifier": "variant2"},
            {"identifier": "variant3"},
        ]
    })

    packages = [
        wiz.package.Package(definition, variant_index=0),
        wiz.package.Package(definition, variant_index=1),
        wiz.package.Package(definition, variant_index=2)
    ]

    nodes = [
        mocker.Mock(identifier="_A[variant1]==0.1.0"),
        mocker.Mock(identifier="_A[variant2]==0.1.0"),
        mocker.Mock(identifier="_A[variant3]==0.1.0"),
    ]

    requirement = Requirement("A")

    mocked_package_extract.return_value = packages

    graph = wiz.graph.Graph(mocked_resolver)
    graph.create_link = mocker.Mock()
    graph._create_node_from_package = mocker.Mock()
    graph._node_mapping = {
        "A[variant1]==0.1.0": nodes[0],
        "A[variant2]==0.1.0": nodes[1],
        "A[variant3]==0.1.0": nodes[2]
    }
    graph.exists = mocker.Mock(return_value=False)

    graph._update_from_requirement(
        requirement, graph.ROOT, mocked_queue, **options
    )

    assert graph._create_node_from_package.call_count == 3
    for package in packages:
        graph._create_node_from_package.assert_any_call(package)

    assert graph.create_link.call_count == 3
    for node in nodes:
        graph.create_link.assert_any_call(
            node.identifier,
            graph.ROOT,
            requirement,
            weight=options.get("weight", 1)
        )

    for node in nodes:
        node.add_parent.assert_called_once_with(graph.ROOT)

    mocked_queue.put.assert_not_called()


def test_graph_update_from_requirements_with_invalid_requirements(
    mocked_resolver, mocked_package_extract
):
    """Fail to update graph with invalid package requirements."""
    graph = wiz.graph.Graph(mocked_resolver)

    _mapping = {
        "A==0.1.0": wiz.package.Package(
            wiz.definition.Definition({
                "identifier": "A",
                "version": "0.1.0",
                "requirements": ["!!!"],
            })
        ),
    }

    mocked_package_extract.side_effect = [
        [_mapping["A==0.1.0"]],
    ]

    with pytest.raises(wiz.exception.IncorrectDefinition) as error:
        graph.update_from_requirements([Requirement("A")], graph.ROOT)

    assert (
        "IncorrectDefinition: Package 'A==0.1.0' is incorrect "
        "[The requirement '!!!' is incorrect]"
    ) in str(error)


def test_graph_update_from_requirements_with_invalid_conditions(
    mocked_resolver, mocked_package_extract
):
    """Fail to update graph with invalid package conditions."""
    graph = wiz.graph.Graph(mocked_resolver)

    _mapping = {
        "A==0.1.0": wiz.package.Package(
            wiz.definition.Definition({
                "identifier": "A",
                "version": "0.1.0",
                "conditions": ["!!!"],
            })
        ),
    }

    mocked_package_extract.side_effect = [
        [_mapping["A==0.1.0"]],
    ]

    with pytest.raises(wiz.exception.IncorrectDefinition) as error:
        graph.update_from_requirements([Requirement("A")], graph.ROOT)

    assert (
        "IncorrectDefinition: Package 'A==0.1.0' is incorrect "
        "[The requirement '!!!' is incorrect]"
    ) in str(error)


def test_graph_create_node_from_package():
    """Create node in graph from package."""
    package = wiz.package.Package(
        wiz.definition.Definition({
            "identifier": "A",
            "version": "0.1.0"
        })
    )

    graph = wiz.graph.Graph(None)
    graph._create_node_from_package(package)

    assert graph._identifiers_per_definition == {"A": {"A==0.1.0"}}
    assert list(graph._node_mapping.keys()) == ["A==0.1.0"]
    assert isinstance(graph._node_mapping["A==0.1.0"], wiz.graph.Node)


@pytest.mark.parametrize("options", [
    {},
    {"weight": 5},
], ids=[
    "simple",
    "with-weight",
])
def test_graph_create_link(options):
    """Create link between two nodes."""
    requirement = Requirement("A")

    graph = wiz.graph.Graph(None)
    graph.create_link("child", "parent", requirement, **options)

    assert graph._link_mapping == {
        "parent": {
            "child": {
                "requirement": requirement,
                "weight": options.get("weight", 1)
            }
        }
    }


def test_graph_create_link_overwrite():
    """Overwrite existing link between two nodes."""
    graph = wiz.graph.Graph(None)
    graph._link_mapping = {
        "parent": {"child": {"requirement": Requirement("A"), "weight": 3}}
    }

    # Ignore when weight is higher
    graph.create_link("child", "parent", Requirement("A>2"), weight=4)

    assert graph.link_requirement("child", "parent") == Requirement("A")
    assert graph.link_weight("child", "parent") == 3

    # Don't ignore when weight is lower
    graph.create_link("child", "parent", Requirement("A>2"), weight=1)

    assert graph.link_requirement("child", "parent") == Requirement("A>2")
    assert graph.link_weight("child", "parent") == 1


def test_graph_remove_node():
    """Remove nodes from graph."""
    graph = wiz.graph.Graph(None)
    graph._node_mapping = {"A1": "_A1", "A2": "_A2", "B": "B"}
    graph._identifiers_per_definition = {"defA": ["A1", "A2"], "defB": ["B"]}
    graph._variants_per_definition = {"_id": ["A1", "A2"]}
    graph._link_mapping = {"A1": {"B": "LINK"}}

    graph.remove_node("A1")

    assert graph._node_mapping == {"A2": "_A2", "B": "B"}
    assert graph._identifiers_per_definition == {
        "defA": ["A1", "A2"], "defB": ["B"]
    }
    assert graph._variants_per_definition == {"_id": ["A1", "A2"]}
    assert graph._link_mapping == {"A1": {"B": "LINK"}}


def test_node():
    """Create and use node."""
    definition = wiz.definition.Definition({
        "identifier": "A",
        "version": "0.1.0",
        "variants": [{"identifier": "V1"}]
    })

    package = wiz.package.Package(definition, variant_index=0)

    node = wiz.graph.Node(package)
    assert node.identifier == "A[V1]==0.1.0"
    assert node.definition == definition
    assert node.package == package
    assert node.parent_identifiers == set()

    node.add_parent("parent1")
    node.add_parent("parent1")
    node.add_parent("parent2")

    assert node.parent_identifiers == {"parent1", "parent2"}


def test_distance_queue():
    """Create and use distance queue."""
    queue = wiz.graph._DistanceQueue()
    assert queue.empty() is True

    # Add element to the queue. This will also be pushed to the heap.
    queue["A"] = 2
    queue["B"] = 1
    queue["C"] = 4
    queue["D"] = 3
    queue["E"] = 6
    queue["F"] = 5
    assert len(queue) == 6
    assert len(queue._heap) == 6
    assert queue.empty() is False

    # Modify all elements. This will override the dictionary items, but
    # it will be pushed to the heap and double its size.
    queue["A"] = 20
    queue["B"] = 10
    queue["C"] = 40
    queue["D"] = 30
    queue["E"] = 60
    queue["F"] = 50
    assert len(queue) == 6
    assert len(queue._heap) == 12

    # By asking the smallest item, half of the heap need to be "popped" as
    # The smallest element is 1 from a previous version of "B". It will then
    # check all elements from the heap until one that also exist in the
    # dictionary is found.
    assert queue.pop_smallest() == "B"
    assert len(queue) == 5
    assert len(queue._heap) == 5

    queue["A"] = 200
    queue["C"] = 400
    queue["D"] = 300
    queue["E"] = 600
    queue["F"] = 500
    assert len(queue) == 5
    assert len(queue._heap) == 10

    # Augment the heap so that it get more than twice the size of the
    # dictionary to force a rebuild.
    queue["A"] = 2000
    assert len(queue) == 5
    assert len(queue._heap) == 5

    # Empty the dictionary.
    assert queue.pop_smallest() == "D"
    assert queue.pop_smallest() == "C"
    assert queue.pop_smallest() == "F"
    assert queue.pop_smallest() == "E"
    assert queue.pop_smallest() == "A"
    assert queue.empty() is True



# :coding: utf-8

import base64
import collections
import hashlib
import pipes
import re
import zlib

import colorama
from packaging.requirements import InvalidRequirement
from packaging.version import Version, InvalidVersion
import ujson

import wiz.exception
import wiz.symbol
from ._requirement import Requirement

# Arbitrary number which indicates a very high version number
_INFINITY_VERSION = 9999


def get_requirement(content):
    """Return the corresponding requirement instance from *content*.

    :param content: String representing a requirement, with or without
        version specifier or variant (e.g. "maya", "nuke >= 10, < 11",
        "ldpk-nuke[10.0]").

    :return: Instance of :class:`packaging.requirements.Requirement`.

    :raise: :exc:`wiz.exception.InvalidRequirement` if the requirement is
        incorrect.

    """
    try:
        return Requirement(content)
    except InvalidRequirement:
        raise wiz.exception.InvalidRequirement(
            "The requirement '{}' is incorrect".format(content)
        )


def get_version(content):
    """Return the corresponding version instance from *content*.

    :param content: String representing a version (e.g. "2018", "0.1.0").

    :return: Instance of :class:`packaging.version.Version`.

    :raise: :exc:`wiz.exception.InvalidVersion` if the version is incorrect.

    """
    try:
        return Version(content)
    except InvalidVersion:
        raise wiz.exception.InvalidVersion(
            "The version '{}' is incorrect".format(content)
        )


def is_overlapping(requirement1, requirement2):
    """Indicate whether requirements are overlapping.

    A requirement is overlapping with another one if their intersection of
    version ranges are not empty.

    Example::

        >>> is_overlapping(Requirement("foo >= 10"), Requirement("foo < 9"))
        True

        >>> is_overlapping(Requirement("foo >= 10"), Requirement("foo < 8"))
        False

    :param requirement1: Instance of
        :class:`packaging.requirements.Requirement`.

    :param requirement2: Instance of
        :class:`packaging.requirements.Requirement`.

    :return: Boolean value.

    :raise: :exc:`wiz.exception.GraphResolutionError` if requirements cannot
        be compared.

    """
    if requirement1.name != requirement2.name:
        raise wiz.exception.GraphResolutionError(
            "Impossible to compare requirements with different names "
            "['{}' and '{}'].".format(requirement1.name, requirement2.name)
        )

    r1 = extract_version_ranges(requirement1)
    r2 = extract_version_ranges(requirement2)
    return (
        (r2[-1][1] is None or r1[0][0] is None or r2[-1][1] >= r1[0][0]) and
        (r1[-1][1] is None or r2[0][0] is None or r1[-1][1] >= r2[0][0])
    )


def extract_version_ranges(requirement):
    """Extract version ranges from *requirement*.

    Requirement could contain the following specifiers:

    * `Compatible release
      <https://www.python.org/dev/peps/pep-0440/#compatible-release>`_
    * `Version matching
      <https://www.python.org/dev/peps/pep-0440/#version-matching>`_
    * `Version exclusion
      <https://www.python.org/dev/peps/pep-0440/#version-exclusion>`_
    * `Inclusive ordered comparison
      <https://www.python.org/dev/peps/pep-0440/#inclusive-ordered-comparison>`_
    * `Exclusive ordered comparison
      <https://www.python.org/dev/peps/pep-0440/#exclusive-ordered-comparison>`_

    Example::

        >>> extract_version_ranges(Requirement("foo"))
        [(None, None)]

        >>> extract_version_ranges(Requirement("foo==2.1.1"))
        [((2, 1, 1), (2, 1, 1))]

        >>> extract_version_ranges(Requirement("foo!=2.1.1"))
        [(None, (2, 1, 0, 9999)), ((2, 1, 1, 1), None)]

        >>> extract_version_ranges(Requirement("foo ==2.*"))
        [((2,), (2, 9999))]

        >>> extract_version_ranges(Requirement("foo >= 2, < 3"))
        [((2,), (2, 9999))]

    :param requirement: Instance of :class:`packaging.requirements.Requirement`.

    :return: List of version tuples.

    :raise: :exc:`wiz.exception.InvalidVersion` if the version extracted from
        the specifier is incorrect.

    :raise: :exc:`wiz.exception.InvalidRequirement` if the specifier operator
        is not accepted or if the requirement does not allow any versions to be
        reached.

    """
    version_ranges = [(None, None)]

    # Sort specifiers for deterministic results.
    for specifier in sorted(
        requirement.specifier, key=lambda s: (s.version, s.operator)
    ):
        # Extract version number.
        _version = wiz.utility.get_version(
            re.sub(r"\.\*$", "", specifier.version)
        ).release

        # Update maximum according to operator.
        if specifier.operator == ">=":
            _update_minimum_version(_version, version_ranges)

        elif specifier.operator == "<=":
            _update_maximum_version(_version, version_ranges)

        elif specifier.operator == ">":
            _min_version = _increment_version(_version)
            _update_minimum_version(_min_version, version_ranges)

        elif specifier.operator == "<":
            _max_version = _decrement_version(_version)
            _update_maximum_version(_max_version, version_ranges)

        elif specifier.operator == "~=":
            _max_version = _increment_version(_version, delta=_INFINITY_VERSION)
            _update_minimum_version(_version, version_ranges)
            _update_maximum_version(_max_version, version_ranges)

        elif specifier.operator == "==" and specifier.version.endswith(".*"):
            _max_version = _increment_version(_version, delta=_INFINITY_VERSION)
            _update_minimum_version(_version, version_ranges)
            _update_maximum_version(_max_version, version_ranges)

        elif specifier.operator == "==":
            _update_minimum_version(_version, version_ranges)
            _update_maximum_version(_version, version_ranges)

        elif specifier.operator == "!=" and specifier.version.endswith(".*"):
            _min_version = _decrement_version(_version)
            _max_version = _increment_version(_version, add_subversion=False)
            _update_version_ranges((_min_version, _max_version), version_ranges)

        elif specifier.operator == "!=":
            _min_version = _decrement_version(_version)
            _max_version = _increment_version(_version)
            _update_version_ranges((_min_version, _max_version), version_ranges)

        else:
            raise wiz.exception.InvalidRequirement(
                "Operator '{}' is not accepted for requirement '{}'".format(
                    specifier.operator, requirement
                )
            )

    return version_ranges


def _update_maximum_version(version, ranges):
    """Update version *ranges* with maximum *version*.

    Example::

        >>> ranges = [((1, 2, 3), (1, 3, 0)), ((1, 3, 3), (1, 4))]
        >>> _update_maximum_version((1, 2, 3), ranges)
        >>> print(ranges)
        [((1, 2, 3), (1, 2, 3))]

    :param version: Tuple representing a version (e.g. (1,2,3)).

    :param ranges: ordered list of tuples containing two ordered version release
        tuples (e.g [((1,2,3), (1,3,0)), (1,3,3), (1,4))]).

    .. warning::

        The *ranges* list will be mutated.

    """
    _ranges = []

    if ranges[0][0] is not None and version < ranges[0][0]:
        raise wiz.exception.InvalidRequirement(
            "The requirement is incorrect as maximum value '{}' cannot be set "
            "when minimum value is '{}'.".format(
                ".".join(str(v) for v in version),
                ".".join(str(v) for v in ranges[0][0])
            )
        )

    for start_version, end_version in ranges:
        if start_version and start_version > version:
            break
        if not end_version or end_version >= version:
            _ranges.append((start_version, version))
            break
        _ranges.append((start_version, end_version))

    # Mutated input ranges
    ranges[:] = _ranges


def _update_minimum_version(version, ranges):
    """Update version *ranges* with minimum *version*.

    Example::

        >>> ranges = [((1, 2, 3), (1, 3, 0)), ((1, 3, 3), (1, 4))]
        >>> _update_minimum_version((1, 3, 3), ranges)
        >>> print(ranges)
        [((1, 3, 3), (1, 4))]

    :param version: Tuple representing a version (e.g. (1,2,3)).

    :param ranges: ordered list of tuples containing two ordered version release
        tuples (e.g [((1,2,3), (1,3,0)), (1,3,3), (1,4))]).

    .. warning::

        The *ranges* list will be mutated.

    """
    _ranges = []

    if ranges[-1][1] is not None and version > ranges[-1][1]:
        raise wiz.exception.InvalidRequirement(
            "The requirement is incorrect as minimum value '{}' cannot be set "
            "when maximum value is '{}'.".format(
                ".".join(str(v) for v in version),
                ".".join(str(v) for v in ranges[-1][1])
            )
        )

    for start_version, end_version in reversed(ranges):
        if end_version and end_version < version:
            break
        if not start_version or start_version <= version:
            _ranges = [(version, end_version)] + _ranges
            break
        _ranges = [(start_version, end_version)] + _ranges

    # Mutated input ranges
    ranges[:] = _ranges


def _update_version_ranges(excluded, ranges):
    """Update version *ranges* from *excluded* version range.

    Example::

        >>> ranges = [((1,2,3), (1,3,0)), ((1,3,3), (1,4))]
        >>> _update_version_ranges(((1,2,3), (1,3,3)), ranges)
        >>> print(ranges)
        [((1, 2, 3), (1, 2, 3)), ((1, 3, 3), (1, 4))]

    :param excluded: Tuple containing two ordered version release tuples (e.g.
        ((1,2,3), (1,3,0))). These two versions are included in *ranges*, but
        all versions in between should be excluded.

    :param ranges: ordered list of tuples containing two ordered version release
        tuples (e.g [((1,2,3), (1,3,0)), (1,3,3), (1,4))]).

    .. warning::

        The *ranges* list will be mutated.

    """
    _ranges = []

    for r in ranges:
        out_before = r[0] is not None and r[0] >= excluded[1]
        out_after = r[1] is not None and r[1] <= excluded[0]

        # Exclusion zone is outside of range.
        if out_before or out_after:
            _ranges.append(r)
            continue

        r0_excluded = r[0] is not None and r[0] > excluded[0]
        r1_excluded = r[1] is not None and r[1] < excluded[1]

        # Exclusion zone covers all range.
        if r0_excluded and r1_excluded:
            continue

        # Exclusion zone cover start of range only.
        elif r0_excluded and not r1_excluded:
            _ranges.append((excluded[1], r[1]))

        # Exclusion zone cover end of range only.
        elif not r0_excluded and r1_excluded:
            _ranges.append((r[0], excluded[0]))

        # Exclusion zone cover middle of range.
        elif not r0_excluded and not r1_excluded:
            _ranges.append((r[0], excluded[0]))
            _ranges.append((excluded[1], r[1]))

    if len(_ranges) == 0:
        raise wiz.exception.InvalidRequirement(
            "The requirement is incorrect as excluded version range '{}-{}' "
            "makes all other versions unreachable.".format(
                ".".join(str(v) for v in excluded[0]),
                ".".join(str(v) for v in excluded[1]),
            )
        )

    # Mutated input ranges
    ranges[:] = _ranges


def _increment_version(version, delta=1, add_subversion=True):
    """Increment *version*.

    This will attempt to increase to the nearest possible version tuple.

    Example::

        >>> _increment_version((1, 2, 0))
        (1, 2, 0, 1)

        >>> _increment_version((1, 2, 0), add_subversion=False)
        (1, 2, 1)

        >>> _increment_version((1, 1, 1), delta=3)
        (1, 1, 1, 3)

    :param version: Tuple representing a version (e.g. (1,2,3)).

    :param delta: Number to add to the minimal the *version*. Default is 1.

    :param add_subversion: Indicate whether a sub-version should be used instead
        of only increasing the minimal release version. Default is True.

    :return: New version tuple.

    """
    if not add_subversion:
        return version[:-1] + (version[-1] + delta,)
    return version + (delta,)


def _decrement_version(version):
    """Decrement *version*.

    This will attempt to decrease to the nearest possible version tuple.

    Example::

        >>> _decrement_version((1,))
        (0, 9999)

        >>> _decrement_version((1, 0, 0))
        (1, 9999)

        >>> _decrement_version((1, 2, 0))
        (1, 1, 9999)

        >>> _decrement_version((1, 1, 1))
        (1, 1, 0, 9999)

    :param version: Tuple representing a version (e.g. (1,2,3)).

    :return: New version tuple.

    """
    index = -1
    while version[index] == 0:
        index -= 1

    return version[:index] + (version[index] - 1, _INFINITY_VERSION)


def encode(element):
    """Return serialized and encoded *element*.

    *element* is serialized first, then encoded into :term:`base64`.

    :param element: Content to encode.

    :raise: :exc:`TypeError` if *element* is not JSON serializable.

    """
    return base64.b64encode(zlib.compress(ujson.dumps(element).encode("utf-8")))


def decode(element):
    """Return deserialized and decoded *element*.

    *element* is decoded first from :term:`base64`, then deserialized.

    :param element: Content to decode.

    :raise: :exc:`TypeError` if *element* cannot be decoded or deserialized.

    """
    return ujson.loads(zlib.decompress(base64.b64decode(element)))


def compute_label(definition):
    """Return unique label for *definition*.

    The name should be in the form of::

        "'foo'"
        "'bar' [0.1.0]"
        "'baz' [0.2.0] (linux : el =! 7)"
        "'bim' (linux : el >= 6, < 7)"

    :param definition: Instance of :class:`wiz.definition.Definition`.

    :return: String representing definition.

    """
    label = "'{}'".format(definition.qualified_identifier)

    if definition.version:
        label += " [{}]".format(definition.version)

    if definition.system:
        system_identifier = compute_system_label(definition)
        label += " ({})".format(system_identifier)

    return label


def compute_system_label(definition):
    """Return unique system label from *definition*.

    The system identifier should be in the form of::

        "noarch"
        "linux : x86_64 : el >= 7, < 8"
        "centos >= 7, < 8"
        "x86_64 : el >= 7, < 8"
        "windows"

    :param definition: Instance of :class:`wiz.definition.Definition`.

    :return: String representing system identifier.

    """
    elements = [
        definition.system.get(element)
        for element in ["platform", "arch", "os"]
        if definition.system.get(element) is not None
    ]
    return " : ".join(elements) or "noarch"


def compute_file_name(definition):
    """Return unique file name from *definition*.

    The file name should be in the form of::

        "foo.json"
        "namespace-foo-0.1.0.json"
        "foo-0.1.0.json"
        "foo-0.1.0-M2Uq9Esezm-m00VeWkTzkQIu3T4.json"

    :param definition: Instance of :class:`wiz.definition.Definition`.

    :return: File name representing definition.

    """
    name = definition.identifier

    if definition.namespace:
        name = "{}-{}".format(
            "-".join(definition.namespace.split(wiz.symbol.NAMESPACE_SEPARATOR)), name
        )

    if definition.version:
        name += "-{}".format(definition.version)

    if definition.system:
        system_identifier = wiz.utility.compute_system_label(definition)
        data = re.sub(r"(\s+|:+)", "", system_identifier).encode("utf-8")
        encoded = base64.urlsafe_b64encode(hashlib.sha1(data).digest())
        name += "-{}".format(encoded.rstrip(b"=").decode("utf-8"))

    return "{}.json".format(name)


def combine_command(elements):
    """Return command *elements* as a string.

    Example::

        >>> combine_command(
        ...     ['python2.7', '-c', 'import os; print(os.environ["HOME"])'])
        ... )

        python2.7 -c 'import os; print(os.environ["HOME"])'

    :param elements: List of strings constituting the command line to execute
        (e.g. ["app_exe", "--option", "value"])


    """
    return " ".join([pipes.quote(element) for element in elements])


def colored_text(message, color):
    """Return colored *message* according to color *name*.

    Available color names are: "black", "red", "green", "yellow", "blue",
    "magenta", "cyan" and "white".

    :param message: String to colorize.

    :param color: Symbol of color to apply to *message*.

    :return: Colorized message.

    """
    return (
        getattr(colorama.Fore, color.upper()) + message +
        colorama.Style.RESET_ALL
    )


def deep_update(mapping1, mapping2):
    """Recursively update *mapping1* from *mapping2*.

    Contrary to :meth:`dict.update`, this function will attempt to update
    sub-dictionaries defined in both mappings instead of overwriting the value
    defined in *mapping1*::

        >>> deep_update({"A": {"B": 2}}, {"A": {"C": 3}})
        {"A": {"B": 2, "C": 3}}

    :param mapping1: Mapping to update

    :param mapping2: Mapping to update *mapping1* from

    :return: *mapping1* mutated.

    .. note::

        *mapping1* will be mutated, but *mapping2* will not.

    """
    for key, value in mapping2.items():
        if isinstance(value, collections.Mapping):
            mapping1[key] = deep_update(mapping1.get(key, {}), value)
        else:
            mapping1[key] = value
    return mapping1
